import pathToRegexp from 'path-to-regexp';

export default class ActivityRouter {

  /**
   * @typedef {Object} Context
   * @param {Element} container
   */

  /**
   * @typedef {Object} Route
   * @param {RegExp} pattern
   * @param {Array<object>} options
   * @param {Activity} activity
   */

  /**
   * @param {Context} context
   */
  constructor(context) {

    /**
     * @property {Array<Route>} currentPath
     */
    this.routes = [];

    /**
     * @property {Context} context
     */
    this.context = context;

    /**
     * @property {String} currentPath
     */
    this.currentPath = null;

    /**
     * @property {WeakMap<Activity, Activity>} createdActivities
     */
    this.createdActivities = new WeakMap();

    /**
     * @property {Activity} errorActivity
     */
    this.errorActivity = null;

    // handling navigation
    window.addEventListener('popstate', this.onPopState.bind(this));
    document.addEventListener('click', this.onClickAnchor.bind(this));
  }

  onPopState(event) {
    event.preventDefault();
    this.updateCurrentActivity();
  }

  onClickAnchor(event) {
    let el = event.target;
    do {
      if (el.tagName === 'A') {
        // without external link
        if (el.href.indexOf(location.host) !== -1) {
          event.preventDefault();
          this.navigateRoute(el.href);
        }
        return;
      }
      el = el.parentNode;
    } while(el !== document.body);
  }

  /**
   * @param {String} href
   */
  navigateRoute(href) {
    history.pushState(undefined, '', href);
    this.updateCurrentActivity();
  }

  navigate404() {
    this.errorActivity.onStart({
      statusCode : 404,
      message    : 'Not Found'
    });
  }

  /**
   * @param {String} pattern
   * @param {Activity} ActivityClass
   */
  addRoute(pattern, ActivityClass) {
    let options = [];
    this.routes.push({
      pattern  : pathToRegexp(pattern, options),
      options  : options,
      activity : ActivityClass
    });
  }

  /**
   * @param {Activity} ErrorActivity
   */
  setErrorActivity(ErrorActivity) {
    if (this.errorActivity) {
      this.errorActivity.onDestroy();
    }
    this.errorActivity = new ErrorActivity(this.context.content);
  }

  updateCurrentActivity() {
    let newPath = document.location.pathname;
    let newActivity = this.resolveActivityByPath(newPath);
    let currentActivity = this.resolveActivityByPath(this.currentPath);

    if (newActivity && newActivity === currentActivity) {
      newActivity.onUpdate(this.resolveParameterByPath(newPath));
      return;
    }

    if (currentActivity) {
      currentActivity.onStop();
    }

    if (newActivity) {
      newActivity.onStart(this.resolveParameterByPath(newPath));
      this.currentPath = newPath;
    } else {
      if (this.errorActivity) {
        this.navigate404();
      } else {
        throw new Error('Unexpected routing path: ' + newPath);
      }
      this.currentPath = null;
    }
  }

  /**
   * @param {String} path
   */
  resolveActivityByPath(path) {
    let matchRoute = this.routes.filter((route) => route.pattern.test(path))[0];
    if (!matchRoute) {
      return null;
    }

    let ActivityClass = matchRoute.activity;
    if (!this.createdActivities.has(ActivityClass)) {
      this.createdActivities.set(ActivityClass, new ActivityClass(this.context));
    }

    return this.createdActivities.get(ActivityClass);
  }

  /**
   * @param {String} path
   */
  resolveParameterByPath(path) {
    let matchRoute = this.routes.filter((route) => route.pattern.test(path))[0];
    if (!matchRoute) {
      return null;
    }

    let matchResult = matchRoute.pattern.exec(path).slice(1);
    let returnValue = matchRoute.options.reduce((rv, option, i) => {
      rv[option.name] = matchResult[i];
      return rv;
    }, {});

    return returnValue;
  }
}
